/**********************************
 * FFILTERS.H
 *
 * Basic Font Filtering Scheme
 *
 * James Bumgardner 1987
 ***********************************/

#include <stdlib.h>
#include <memory.h>

#if defined(HUGEC)
#define SIZE_LIMIT 300
#define BUFLIM 8000
#define BIG huge
#else
#define SIZE_LIMIT 152
#define BUFLIM 3000
#define BIG
#endif

#define ESC 27

#define TWO_CHARS(c1, c2) (((c2) << 8) + (c1))

struct {
	int fil26;
	char nul1,font_type;
	int nul2,base_line,width,height;
	char orient,fixed;
	int symset,pitch,points,nul3;
	char nul4,style,weight,type_face;
} font_header;

struct {
	char fil4,nul1,fil14,cil1;
	char orient,nul2;
	int left_offset,top_offset,width,height,delta_x;
} char_header;

char ibuf[BUFLIM],*bitmap = &ibuf[sizeof(char_header)];

char BIG pmap[2][SIZE_LIMIT][SIZE_LIMIT];

FILE *ifile,*ofile;

void erxit(char *msg);
void erxit(msg)
char *msg;
{
	fprintf(stderr,msg);
	exit(-1);
}

int get_number(void);
int get_number()
{
	char *p;
	int c;

	p = ibuf;
	while ((c = getc(ifile)) != EOF && isdigit(c))
		*p++ = (char) c;
	*p = '\0';
	return atoi(ibuf);
}

int iswap(unsigned int n);
int iswap(unsigned int n)
{
	return(n >> 8 | n << 8);
}

int get_font_header(void);
int get_font_header()
{
	unsigned int tmp;
	int c;
	char *p;
	
	while ((c = getc(ifile)) != EOF && c != ESC)
		;
	if (c == EOF)
		erxit("Invalid Font File");
	tmp = getw(ifile);

	if (tmp == TWO_CHARS(')', 's')	) {
		tmp = get_number();
		if (tmp > sizeof(ibuf))
			erxit("Header too large for input buffer!\n");
		fread(ibuf, tmp, 1, ifile);
		memcpy(&font_header,ibuf,sizeof(font_header));
		font_header.base_line = iswap(font_header.base_line);
		font_header.width = iswap(font_header.width);
		font_header.height = iswap(font_header.height);
		if (font_header.width > SIZE_LIMIT ||
			font_header.height > SIZE_LIMIT) {
			fprintf(stderr,"Characters are too big [%d x %d]\n",
				font_header.width, font_header.height);
			return EOF;
		}
		return tmp;
	}
	return EOF;
}

void write_font_header(int t);
void write_font_header(int t)
{
	fprintf(ofile,"\033)s%dW",t);
	font_header.base_line = iswap(font_header.base_line);
	font_header.width = iswap(font_header.width);
	font_header.height = iswap(font_header.height);
	memcpy(ibuf,&font_header,sizeof(font_header));
	fwrite(ibuf,t,1,ofile);
}

int get_char_header(void);
int get_char_header()
{
	int tmp;
	int c,cno;

	while ((c = getc(ifile)) != EOF && c != ESC)
		;
	if (c == EOF)
		return EOF;
	tmp = getw(ifile);

	if (tmp != TWO_CHARS('*', 'c'))
		return EOF;

	cno = get_number();

	printf("\b%c",cno);
	fprintf(ofile,"\033*c%dE",cno);

	c = getc(ifile);		/* skip escape */
	tmp = getw(ifile);
	if (tmp != TWO_CHARS('(', 's'))
		return EOF;

	tmp = get_number();
	if (tmp > sizeof(ibuf))
		erxit("Overflow input buffer!\n");

	fread(ibuf,tmp,1,ifile);
	return tmp;
}

void fix_header(void);
void fix_header()
{
	memcpy(&char_header,ibuf,sizeof(char_header));
	char_header.width = iswap(char_header.width);
	char_header.height = iswap(char_header.height);
	char_header.delta_x = iswap(char_header.delta_x);
	char_header.left_offset = iswap(char_header.left_offset);
	char_header.top_offset = iswap(char_header.top_offset);
}

void unfix_header(void);
void unfix_header()
{
	char_header.width = iswap(char_header.width);
	char_header.height = iswap(char_header.height);
	char_header.delta_x = iswap(char_header.delta_x);
	char_header.left_offset = iswap(char_header.left_offset);
	char_header.top_offset = iswap(char_header.top_offset);
	memcpy(ibuf,&char_header,sizeof(char_header));
}

void translate_bitmap(int h, int w);
void translate_bitmap(int h, int w)
{
	int y,x;
	char *p;
	char BIG *op;
	unsigned char byt,bit;

	p = bitmap;
	for (y = 0; y < h; ++y) {
		op = pmap[0][y];
		for (x = 0; x < w; ++x) {
			byt = *p++;
			for (bit = 0x80; bit; bit >>= 1) {
				if (byt & bit)
					*op++ = 1;
				else
					*op++ = 0;
			}
		}
	}
}

void translate_pmap(int h, int w);
void translate_pmap(int h, int w)
{
	int y,x,rx;
	char *p;
	unsigned char byt,bit;
	p = bitmap;
	for (y = 0; y < h; ++y) {
		rx = 0;
		for (x = 0; x < w; ++x) {
			byt = 0;
			for (bit = 0x80; bit; bit >>= 1) {
				if (pmap[1][y][rx])
					byt |= bit;
				++rx;
			}
			*p++ = byt;
		}

	}
}

main(argc,argv)
char *argv[];
{
	int total_bytes, horiz_bytes, true_width, arg;

	if (argc < 3)
		erxit("Please Specify an Input and Output File");

	if (argc >= 4)
		arg = atoi(argv[3]);
	else
		arg = 0;	
	
	if ((ifile = fopen(argv[1],"rb")) == NULL)
		erxit("Can't open Input File\n");
	if ((ofile = fopen(argv[2],"wb")) == NULL)
		erxit("Can't open Output File\n");

	if ((total_bytes = get_font_header()) == EOF)
		erxit("Invalid Font File");
	process_font_header(arg);
	write_font_header(total_bytes);
		
	printf("Processing  ");
	while ((total_bytes = get_char_header()) != EOF)
	{
		fix_header();

		if (char_header.height != 0) {
			horiz_bytes = (total_bytes-sizeof(char_header)) /
				char_header.height;

			translate_bitmap(char_header.height,horiz_bytes);

			true_width = horiz_bytes << 3;
			process_char(char_header.height,true_width,arg);

			horiz_bytes = (char_header.width-1 >> 3)+1;
			total_bytes = sizeof(char_header)+horiz_bytes*char_header.height;
			if (total_bytes > sizeof(ibuf))
				erxit("New character > output buffer\n");
			translate_pmap(char_header.height,horiz_bytes);

		}

		unfix_header();

		fprintf(ofile,"\x01B(s%dW",total_bytes);
		fwrite(ibuf,total_bytes,1,ofile);
	}
}
